from asgikit.requests import HttpRequest
from asgikit.websockets import WebSocket

__all__ = ("RequestContext",)


class RequestContext:
    __slots__ = ("attributes", "_inner", "__weakref__")

    def __init__(self, scope, receive, send):
        if scope["type"] == "http":
            self._inner = HttpRequest(scope, receive, send)
        elif scope["type"] == "websocket":
            self._inner = WebSocket(scope, receive, send)
        else:
            raise RuntimeError(
                f"scope[type] is neither http nor websocket ({scope['type']})"
            )

        self.attributes = {}

    @property
    def request(self):
        return self._inner if self._inner.is_http else None

    @property
    def websocket(self):
        return self._inner if self._inner.is_websocket else None

    def __getattr__(self, item):
        return getattr(self._inner, item)

    def __getitem__(self, item):
        return self.attributes[item]

    def __setitem__(self, key, value):
        self.attributes[key] = value



import inspect
from enum import Enum
from typing import Callable

from asgikit.requests import HttpMethod

CONTROLLER_ATTRIBUTE = "__selva_web_controller__"
ACTION_ATTRIBUTE = "__selva_web_action__"
PATH_ATTRIBUTE = "__selva_web_path__"


class ActionType(Enum):
    GET = HttpMethod.GET
    POST = HttpMethod.POST
    PUT = HttpMethod.PUT
    PATCH = HttpMethod.PATCH
    DELETE = HttpMethod.DELETE
    WEBSOCKET = None

    @property
    def is_websocket(self):
        return self is ActionType.WEBSOCKET


def controller(path_or_cls: str | type):
    if isinstance(path_or_cls, str):
        path = path_or_cls.strip("/")
        cls = None
    elif inspect.isclass(path_or_cls):
        path = ""
        cls = path_or_cls
    else:
        raise ValueError(f"@controller must be applied to class, '{path_or_cls}' given")

    def inner(arg: type):
        if not inspect.isclass(arg):
            raise ValueError(f"@controller must be applied to class, '{arg}' given")

        setattr(arg, CONTROLLER_ATTRIBUTE, True)
        setattr(arg, PATH_ATTRIBUTE, path)
        return arg

    return inner(cls) if cls else inner


def route(action=None, *, method: HttpMethod | None, path: str | None):
    path = path.strip("/") if path else ""

    def inner(arg: Callable):
        setattr(arg, ACTION_ATTRIBUTE, ActionType(method))
        setattr(arg, PATH_ATTRIBUTE, path)
        return arg

    return inner(action) if action else inner


def _route(method: HttpMethod | None, path_or_action: str | Callable):
    if isinstance(path_or_action, str):
        path = path_or_action.strip("/")
        action = None
    else:
        path = ""
        action = path_or_action

    return route(action, method=method, path=path)


def get(path_or_action: str | type):
    return _route(HttpMethod.GET, path_or_action)


def post(path_or_action: str | type):
    return _route(HttpMethod.POST, path_or_action)


def put(path_or_action: str | type):
    return _route(HttpMethod.PUT, path_or_action)


def patch(path_or_action: str | type):
    return _route(HttpMethod.PATCH, path_or_action)


def delete(path_or_action: str | type):
    return _route(HttpMethod.DELETE, path_or_action)


def websocket(path_or_action: str | type):
    return _route(None, path_or_action)



import inspect
from collections import OrderedDict
from typing import Optional

from asgikit.requests import HttpMethod, HttpRequest
from asgikit.websockets import WebSocket

from selva.web.errors import NotFoundError

from .decorators import ACTION_ATTRIBUTE, CONTROLLER_ATTRIBUTE, PATH_ATTRIBUTE
from .route import Route, RouteMatch


def _name_transform(name: str) -> str:
    yield name[0].lower()
    for letter in name[1:]:
        if letter.isupper():
            yield "-"
            yield letter.lower()
        else:
            yield letter


def _path_with_prefix(path: str, prefix: str):
    path = path.strip("/")
    prefix = prefix.strip("/")
    return f"{prefix}/{path}"


class Router:
    def __init__(self):
        self.routes: OrderedDict[str, Route] = OrderedDict()

    def route(self, controller: type):
        if not hasattr(controller, CONTROLLER_ATTRIBUTE):
            raise ValueError(
                f"{controller.__module__}.{controller.__qualname__}"
                " is not annotated with @controller"
            )

        path_prefix = getattr(controller, PATH_ATTRIBUTE)

        for name, action in inspect.getmembers(controller, inspect.isfunction):
            action_type = getattr(action, ACTION_ATTRIBUTE, None)
            if not action_type:
                continue

            method = action_type.value
            path = getattr(action, PATH_ATTRIBUTE)
            path = _path_with_prefix(path, path_prefix)
            route_name = f"{controller.__module__}.{controller.__qualname__}:{name}"

            route = Route(method, path, controller, action, route_name)

            if (
                action_type.is_websocket
                and HttpRequest in route.request_params.values()
            ):
                raise TypeError("websocket route cannot receive HttpRequest")

            if (
                not action_type.is_websocket
                and WebSocket in route.request_params.values()
            ):
                raise TypeError("http route cannot receive WebSocket")

            for current_route in self.routes.values():
                if (
                    current_route.method == route.method
                    and current_route.regex == route.regex
                ):
                    raise ValueError(
                        f"path for {route.name} clashes with {current_route.name}"
                    )

            self.routes[route_name] = route

    def match(self, method: HttpMethod | None, path: str) -> Optional[RouteMatch]:
        for route in self.routes.values():
            if (match := route.match(method, path)) is not None:
                return RouteMatch(route, method, path, match)

        return None

    def reverse(self, name: str, **kwargs) -> str:
        if route := self.routes.get(name):
            return route.reverse(**kwargs)

        raise NotFoundError()



import asyncio
import functools
import inspect
import logging
from importlib import import_module
from collections.abc import Callable
from http import HTTPStatus
from types import ModuleType
from typing import Any, Optional, Type, TypeGuard

from asgikit.responses import HttpResponse

from selva.di import Container
from selva.di.decorators import DI_SERVICE_ATTRIBUTE
from selva.utils.base_types import get_base_types
from selva.utils.maybe_async import maybe_async
from selva.utils.package_scan import scan_packages
from selva.web.configuration import Settings
from selva.web.errors import HttpError
from selva.web.middleware.base import Middleware
from selva.web.middleware.decorators import MIDDLEWARE_ATTRIBUTE
from selva.web.request import FromRequest, RequestContext, from_request_impl
from selva.web.response import IntoResponse, into_response_impl
from selva.web.routing import param_converter
from selva.web.routing.decorators import CONTROLLER_ATTRIBUTE
from selva.web.routing.router import Router

LOGGER = logging.getLogger(__name__)


def _is_controller(arg) -> bool:
    return inspect.isclass(arg) and hasattr(arg, CONTROLLER_ATTRIBUTE)


def _is_service(arg) -> bool:
    return hasattr(arg, DI_SERVICE_ATTRIBUTE)


def _is_middleware(arg) -> TypeGuard[Type[Middleware]]:
    return hasattr(arg, MIDDLEWARE_ATTRIBUTE)


def _is_registerable(arg) -> bool:
    return any(i(arg) for i in [_is_controller, _is_service])


background_tasks = set()


class Application:
    """Entrypoint class for a Selva Application

    Will try to automatically import and register a application called "application".
    Other modules and classes can be registered using the "register" method
    """

    def __init__(
        self,
        *components,
        middleware: list[Type[Middleware]] = None,
        debug=False,
    ):
        self.debug = debug
        self.di = Container()
        self.router = Router()
        self.handler = self._process_request
        self.middleware_classes: list[Type[Middleware]] = middleware or []

        self.di.define_singleton(Router, self.router)
        self.di.scan(from_request_impl, param_converter, into_response_impl)

        self.di.define_singleton(Settings, Settings())

        for mid in self.middleware_classes:
            self.di.register(mid)

        components = set(components)

        # try to automatically import and register a application called "application"
        try:
            import application as app
        except ImportError:
            app = None

        if app and (app not in components or app.__name__ not in components):
            components.add(app)

        self.register(*components)

    async def __call__(self, scope, receive, send):
        match scope["type"]:
            case "http" | "websocket":
                await self._handle_request(RequestContext(scope, receive, send))
            case "lifespan":
                await self._handle_lifespan(scope, receive, send)
            case _:
                raise RuntimeError(f"unknown scope '{scope['type']}'")

    def register(self, *args: type | Callable | Type[Middleware] | ModuleType | str):
        for item in args:
            if _is_controller(item):
                self.router.route(item)
            elif _is_service(item):
                self.di.service(item)
            elif inspect.ismodule(item) or isinstance(item, str):
                for subitem in scan_packages([item], _is_registerable):
                    if _is_controller(subitem):
                        self.router.route(subitem)
                    elif _is_service(subitem):
                        self.di.service(subitem)
            else:
                raise ValueError(f"{item} is not a controller, service or application")

    async def _handle_lifespan(self, _scope, receive, send):
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                try:
                    await send({"type": "lifespan.startup.complete"})
                except Exception as err:
                    await send({"type": "lifespan.startup.failed", "message": str(err)})
            elif message["type"] == "lifespan.shutdown":
                try:
                    await self.di.run_finalizers()
                    await send({"type": "lifespan.shutdown.complete"})
                except Exception as err:
                    await send(
                        {"type": "lifespan.shutdown.failed", "message": str(err)}
                    )
                break

    async def _params_from_request(
        self,
        context: RequestContext,
        params: dict[str, type],
    ) -> dict[str, Any]:
        request_params = {}

        for name, item_type in params.items():
            for base_type in get_base_types(item_type):
                if converter := await self.di.get(
                    FromRequest[base_type], optional=True
                ):
                    value = await maybe_async(converter.from_request, context)
                    request_params[name] = value
                    break
            else:
                raise RuntimeError(
                    f"no implementation of 'FromRequest' found for type {item_type}"
                )

        return request_params

    async def _handle_request(self, context: RequestContext):
        for cls in self.middleware_classes:
            middleware = await self.di.get(cls)
            self.handler = functools.partial(middleware.execute, self.handler)

        Application._handle_request = Application._initialized_handle_request

        return await self._initialized_handle_request(context)

    async def _initialized_handle_request(self, context: RequestContext):
        try:
            if response := await self.handler(context):
                await response(context.request)
        finally:
            task = asyncio.create_task(self.di.run_finalizers(context))
            background_tasks.add(task)
            task.add_done_callback(background_tasks.discard)

    async def _process_request(self, context: RequestContext) -> HttpResponse:
        self.di.define_dependent(RequestContext, context, context=context)

        method = context.method if context.is_http else None
        path = context.path

        match = self.router.match(method, path)

        if not match:
            return HttpResponse(status=HTTPStatus.NOT_FOUND)

        controller = match.route.controller
        action = match.route.action
        path_params = match.params

        try:
            request_params = await self._params_from_request(
                context, match.route.request_params
            )

            all_params = path_params | request_params
            instance = await self.di.create(controller, context=context)
            response = await maybe_async(action, instance, **all_params)

            if context.is_http:
                return await self._into_response(response)
        except HttpError as err:
            return HttpResponse(status=err.status)
        except Exception as err:
            LOGGER.exception(err)
            return HttpResponse(status=HTTPStatus.INTERNAL_SERVER_ERROR)

    async def _into_response(self, value: Any | None) -> Optional[HttpResponse]:
        if isinstance(value, HttpResponse):
            return value

        for base in get_base_types(type(value)):
            if converter := await self.di.get(IntoResponse[base], optional=True):
                return await maybe_async(converter.into_response, value)

        raise RuntimeError(
            f"no implementation of 'IntoResponse' found for type '{type(value)}'"
        )



import inspect
import typing
import warnings
from collections.abc import Callable
from typing import Annotated, Any, Optional, TypeVar, Union

from selva.di.annotations import Name
from selva.di.decorators import DI_FINALIZER_ATTRIBUTE, DI_INITIALIZER_ATTRIBUTE
from selva.di.errors import (
    FactoryMissingReturnTypeError,
    IncompatibleTypesError,
    InvalidServiceTypeError,
    MultipleNameAnnotationsError,
    NonInjectableTypeError,
    TypeVarInGenericServiceError,
)
from selva.di.service.model import (
    InjectableType,
    Scope,
    ServiceDefinition,
    ServiceDependency,
)


def _get_optional(type_hint) -> tuple[type, bool]:
    if typing.get_origin(type_hint) is Union:
        type_arg = typing.get_args(type_hint)[0]
        if type_hint == Optional[type_arg]:
            return type_arg, True
    return type_hint, False


def _get_annotations(type_hint) -> tuple[type, list[Any]]:
    if typing.get_origin(type_hint) is Annotated:
        type_hint, *annotations = typing.get_args(type_hint)
        return type_hint, annotations
    return type_hint, []


def get_dependencies(service: InjectableType) -> list[tuple[str, ServiceDependency]]:
    if inspect.isclass(service):
        types = typing.get_type_hints(service.__init__, include_extras=True)
    elif inspect.isfunction(service) or inspect.ismethod(service):
        types = typing.get_type_hints(service, include_extras=True)
    elif call := getattr(service, "__call__", None):
        types = typing.get_type_hints(call, include_extras=True)
    else:
        raise InvalidServiceTypeError(service)

    types.pop("return", None)

    result = []

    for name, type_hint in types.items():
        type_hint, optional = _get_optional(type_hint)
        type_hint, annotations = _get_annotations(type_hint)

        # in case Optional is wrapped in Annotated
        if not optional:
            type_hint, optional = _get_optional(type_hint)

        dependency_names = [a.value for a in annotations if isinstance(a, Name)]
        if len(dependency_names) > 1:
            raise MultipleNameAnnotationsError(dependency_names, name, service)
        dependency_name = dependency_names[0] if len(dependency_names) == 1 else None

        service_dependency = ServiceDependency(
            type_hint, optional=optional, name=dependency_name
        )
        result.append((name, service_dependency))

    return result


def _parse_definition_class(
    service: type, provides: Optional[type]
) -> tuple[type, Optional[Callable], Optional[Callable]]:
    initializer = None
    finalizer = None

    if provides:
        origin = typing.get_origin(provides)
        if origin:
            if any(isinstance(a, TypeVar) for a in typing.get_args(provides)):
                raise TypeVarInGenericServiceError(provides)

            orig_bases = getattr(service, "__orig_bases__", None)
            if orig_bases and provides not in orig_bases:
                raise IncompatibleTypesError(service, provides)

        if not issubclass(service, origin or provides):
            raise IncompatibleTypesError(service, provides)

    provided_service = provides or service

    # get initializer and finalizer for service class
    for _, function in inspect.getmembers(service, inspect.isfunction):
        if initializer is not None and finalizer is not None:
            break
        if initializer is None and getattr(function, DI_INITIALIZER_ATTRIBUTE, False):
            initializer = function
        if finalizer is None and getattr(function, DI_FINALIZER_ATTRIBUTE, False):
            finalizer = function

    return provided_service, initializer, finalizer


def _parse_definition_factory(
    service: Callable, provides: Optional[type]
) -> tuple[type, Optional[Callable], Optional[Callable]]:
    initializer = None
    finalizer = None

    if provides:
        message = "option 'provides' on a factory function has no effect"
        warnings.warn(message)

    service_type = typing.get_type_hints(service).get("return")
    if service_type is None:
        raise FactoryMissingReturnTypeError(service)

    provided_service = service_type

    # get initializer and finalizer for service factory
    if func := getattr(service, DI_INITIALIZER_ATTRIBUTE, None):
        initializer = func
    if func := getattr(service, DI_FINALIZER_ATTRIBUTE, None):
        finalizer = func

    return provided_service, initializer, finalizer


def parse_definition(
    service: InjectableType, scope: Scope, provides: type = None
) -> ServiceDefinition:
    if inspect.isclass(service):
        provided_service, initializer, finalizer = _parse_definition_class(
            service, provides
        )
    elif callable(service):
        provided_service, initializer, finalizer = _parse_definition_factory(
            service, provides
        )
    else:
        raise NonInjectableTypeError(service)

    dependencies = get_dependencies(service)

    return ServiceDefinition(
        provides=provided_service,
        scope=scope,
        factory=service,
        dependencies=dependencies,
        initializer=initializer,
        finalizer=finalizer,
    )



import re
import typing
from abc import ABC
from collections import Counter
from re import Pattern
from typing import Callable, NamedTuple

from asgikit.requests import HttpMethod

PATH_PARAMS_RE = re.compile(r"\{([a-zA-Z\w]+)\}")

PATH_PARAM_INT = r"\d+"
PATH_PARAM_FLOAT = r"\d+(\.\d+)?"
PATH_PARAM_BOOL = r"1|0|true|false|True|False"
PATH_PARAM_STR = r".+?"


def build_path_regex(
    path: str, type_hints: dict[str, type]
) -> tuple[dict[str, type], Pattern]:
    """parse path and build regex for route matching

    :returns: tuple of mapping param name to type and compiled regex
    """
    regex = path

    path_params = PATH_PARAMS_RE.findall(path)

    # verify that path does not have duplicate parameters
    counter = Counter(path_params)
    for name, occurencies in counter.items():
        if occurencies > 1:
            raise ValueError(
                f"path parameter '{name}' occurred more than once ({occurencies} times)"
            )

    param_types = {}
    for param in path_params:
        type_hint = type_hints.get(param, str)
        param_types[param] = type_hint

        if type_hint is int:
            param_regex = PATH_PARAM_INT
        elif type_hint is float:
            param_regex = PATH_PARAM_FLOAT
        elif type_hint is bool:
            param_regex = PATH_PARAM_BOOL
        else:
            param_regex = PATH_PARAM_STR

        param_regex = f"(?P<{param}>{param_regex})"
        regex = regex.replace(f"{{{param}}}", param_regex)

    if not regex.startswith("/"):
        regex = "/" + regex

    if regex.endswith("/"):
        regex += "?"
    else:
        regex += "/?"

    regex = f"^{regex}$"
    return param_types, re.compile(regex)


class Route(ABC):
    def __init__(
        self,
        method: HttpMethod | None,
        path: str,
        controller: type,
        action: Callable,
        name: str,
    ):
        self.method = method
        self.path = path
        self.controller = controller
        self.action = action
        self.name = name

        type_hints = typing.get_type_hints(action)
        type_hints.pop("return", None)

        self.path_params, self.regex = build_path_regex(self.path, type_hints)
        self.request_params = {
            name: param_type
            for name, param_type in type_hints.items()
            if name not in self.path_params
        }

    def match(self, method: HttpMethod | None, path: str) -> dict[str, str] | None:
        if method is self.method and (match := self.regex.match(path)):
            return match.groupdict()

        return None

    def reverse(self, **kwargs) -> str:
        path = self.path

        for param, value in kwargs.items():
            if param not in self.path_params:
                raise ValueError()
            path = path.replace(f"{{{param}}}", value)

        return path


class RouteMatch(NamedTuple):
    route: Route
    method: HttpMethod | None
    path: str
    params: dict[str, str]



from collections.abc import Callable
from enum import IntEnum
from types import FunctionType, MethodType
from typing import NamedTuple, Optional

InjectableType = type | FunctionType | MethodType


class Scope(IntEnum):
    SINGLETON = 1
    DEPENDENT = 2
    TRANSIENT = 3


class ServiceInfo(NamedTuple):
    scope: Scope
    provides: Optional[type]
    name: Optional[str]


class ServiceDependency(NamedTuple):
    service: type
    name: Optional[str]
    optional: Optional[bool] = False


class ServiceDefinition(NamedTuple):
    provides: type
    scope: Scope
    factory: InjectableType
    dependencies: list[tuple[str, ServiceDependency]]
    initializer: Optional[Callable]
    finalizer: Optional[Callable]

    def accept_scope(self, scope: Scope) -> bool:
        return self.scope <= scope



import os
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional

try:
    import tomllib as toml
except ImportError:
    import tomli as toml


RESOURCES_DIR_NAME = "resources"
MAIN_SETTINGS_FILE = "application.toml"
ENV_SETTINGS_FILE = "application-{}.toml"
SELVA_ENV = "SELVA_ENV"
SELVA_ENV_PREFIX = "SELVA_"


def _flatten_dict(source: dict) -> Iterable[tuple[str, Any]]:
    for name, value in source.items():
        if isinstance(value, dict):
            for subname, subvalue in _flatten_dict(value):
                yield f"{name}:{subname}".lower(), subvalue
        else:
            yield name.lower(), value


def flatten_dict(source: dict) -> dict[str, Any]:
    return dict(_flatten_dict(source))


def load_config_file(settings_path: Path) -> dict[str, Any]:
    settings = {}

    if settings_path.exists():
        with settings_path.open("rb") as conf:
            settings = toml.load(conf)

        if include := settings.pop("include", None):
            if isinstance(include, str):
                include = [include]

            if isinstance(include, (list, tuple, set)):
                for include_file in include:
                    include_path = settings_path.parent / include_file
                    include_path = include_path.with_suffix(".toml")

                    if not include_path.exists():
                        raise ValueError(f"{include_path} does not exist")

                    with include_path.open("rb") as conf:
                        settings |= toml.load(conf)
            else:
                raise ValueError("'include' must be string or list")

    return flatten_dict(settings)


def load_config_env() -> dict[str, Any]:
    settings = {}

    for key, value in os.environ.items():
        if not key.startswith("SELVA_"):
            continue

        name = key.removeprefix("SELVA_").lstrip("_").lower().replace("__", ":")
        settings[name] = value

    return settings


def get_settings() -> dict[str, Any]:
    resources_dir = Path(os.getcwd()) / RESOURCES_DIR_NAME
    main_settings_file = resources_dir / MAIN_SETTINGS_FILE

    settings = load_config_file(main_settings_file)

    if selva_env := os.getenv(SELVA_ENV):
        env_settings_file = resources_dir / ENV_SETTINGS_FILE.format(selva_env)
        settings |= load_config_file(env_settings_file)

    settings |= load_config_env()

    return settings


class Settings:
    def __init__(self, initial: dict = None):
        self._data = get_settings()
        if initial:
            self._data |= flatten_dict(initial)

    def get(self, *args) -> Optional[Any]:
        key = ":".join(args).lower()
        return self._data.get(key)

    def __getitem__(self, item):
        if isinstance(item, tuple):
            return self.get(*item)
        return self.get(item)

    @property
    def resources_path(self):
        return Path(os.getcwd()) / "resources"

    def resource_path(self, *args: str):
        resources = self.resources_path
        resources.joinpath(*args)



import functools
import inspect
from collections.abc import Callable

from .service.model import InjectableType, Scope, ServiceInfo

DI_SERVICE_ATTRIBUTE = "__selva_di_service__"
DI_INITIALIZER_ATTRIBUTE = "__selva_di_initializer__"
DI_FINALIZER_ATTRIBUTE = "__selva_di_finalizer__"


def service(
    injectable: InjectableType = None,
    *,
    scope: Scope = Scope.SINGLETON,
    provides: type = None,
    name: str = None
):
    def inner(arg: InjectableType):
        setattr(arg, DI_SERVICE_ATTRIBUTE, ServiceInfo(scope, provides, name))
        return arg

    return inner(injectable) if injectable else inner


def initializer(func: Callable):
    setattr(func, DI_INITIALIZER_ATTRIBUTE, True)
    return func


def finalizer(func: Callable):
    if "self" in inspect.signature(func).parameters:
        setattr(func, DI_FINALIZER_ATTRIBUTE, True)
        return func

    @functools.wraps(func)
    def inner(target: Callable):
        setattr(target, DI_FINALIZER_ATTRIBUTE, func)
        return target

    return inner



from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar

__all__ = ("PathParamConverter",)

T = TypeVar("T")


class PathParamConverter(Generic[T], metaclass=ABCMeta):
    @abstractmethod
    async def from_path(self, value: str) -> T:
        pass

    @abstractmethod
    async def to_path(self, obj: T) -> str:
        pass



import inspect

MIDDLEWARE_ATTRIBUTE = "__selva_web_middleware__"


def middleware(target: type = None):
    def inner(arg):
        if not inspect.isclass(arg):
            raise ValueError("Middleware must be a class")

        setattr(arg, MIDDLEWARE_ATTRIBUTE, True)
        return arg

    return inner(target) if target else inner



from asgikit.headers import Headers
from asgikit.query import Query

from selva.di import service

from .context import RequestContext
from .from_request import FromRequest


@service(provides=FromRequest[RequestContext])
class ContextFromRequest(FromRequest[RequestContext]):
    async def from_request(self, context: RequestContext) -> RequestContext:
        return context


@service(provides=FromRequest[Headers])
class HeadersFromRequest(FromRequest[Headers]):
    async def from_request(self, context: RequestContext) -> Headers:
        return context.headers


@service(provides=FromRequest[Query])
class QueryFromRequest(FromRequest[Query]):
    async def from_request(self, context: RequestContext) -> Headers:
        return context.query



import typing
import warnings
from typing import Optional, Union

from ..errors import ServiceAlreadyRegisteredError, ServiceNotFoundError
from .model import ServiceDefinition


class ServiceRecord:
    def __init__(self):
        self.providers: dict[Optional[str], ServiceDefinition] = {}

    def add(self, service: ServiceDefinition, name: str = None):
        if name in self.providers:
            raise ServiceAlreadyRegisteredError(service.provides, name)

        self.providers[name] = service

    def get(self, name: str = None) -> Optional[ServiceDefinition]:
        if service := self.providers.get(name):
            return service

        if default := self.providers.get(None):
            message = (
                f"using default service instead of '{name}'"
                f" for '{default.provides.__qualname__}'"
            )

            warnings.warn(message)
            return default

        return None

    def __contains__(self, name: Optional[str]) -> bool:
        return name in self.providers


def _get_key_with_name(key: type | tuple[type, str]) -> tuple[type, Optional[str]]:
    # typing.get_origin(k) to test Generic[T]
    if isinstance(key, type) or typing.get_origin(key) is not None:
        return key, None

    if isinstance(key, tuple) and len(key) == 2:
        return key

    raise ValueError(key)


class ServiceRegistry:
    def __init__(self):
        self.data: dict[type, ServiceRecord] = {}

    def get(self, key: type, name: str = None) -> Optional[ServiceDefinition]:
        if (key, name) not in self:
            return None
        return self[key, name]

    def __getitem__(self, key: Union[type, tuple[type, str]]):
        inner_key, name = _get_key_with_name(key)

        if (
            inner_key in self.data
            and (service := self.data[inner_key].get(name)) is not None
        ):
            return service

        raise ServiceNotFoundError(inner_key, name=name)

    def __setitem__(self, key: Union[type, tuple[type, str]], value: ServiceDefinition):
        inner_key, name = _get_key_with_name(key)

        if inner_key not in self.data:
            self.data[inner_key] = ServiceRecord()

        self.data[inner_key].add(value, name)

    def __contains__(self, key: Union[type, tuple[type, str]]):
        inner_key, name = _get_key_with_name(key)
        return inner_key in self.data and name in self.data[inner_key]



from asgikit.responses import (
    FileResponse,
    HttpResponse,
    JsonResponse,
    PlainTextResponse,
    RedirectPostGetResponse,
    RedirectResponse,
    SameSitePolicy,
    StreamingResponse,
)
from asgikit.websockets import WebSocket

from selva.web.application import Application
from selva.web.middleware.decorators import middleware
from selva.web.request import RequestContext
from selva.web.routing.decorators import (
    controller,
    delete,
    get,
    patch,
    post,
    put,
    websocket,
)



from http import HTTPStatus

from asgikit.responses import HttpResponse, JsonResponse, PlainTextResponse

from selva.di import service

from .into_response import IntoResponse


@service(provides=IntoResponse[int])
class IntIntoResponse(IntoResponse[int]):
    def into_response(self, value: int) -> HttpResponse:
        return HttpResponse(status=value)


@service(provides=IntoResponse[HTTPStatus])
class HTTPStatusIntoResponse(IntoResponse[HTTPStatus]):
    def into_response(self, value: HTTPStatus) -> HttpResponse:
        return HttpResponse(status=value)


@service(provides=IntoResponse[str])
class StrIntoResponse(IntoResponse[str]):
    def into_response(self, value: str) -> HttpResponse:
        return PlainTextResponse(value)


@service(provides=IntoResponse[list])
class ListIntoResponse(IntoResponse[list]):
    def into_response(self, value: list) -> HttpResponse:
        return JsonResponse(value)


@service(provides=IntoResponse[dict])
class DictIntoResponse(IntoResponse[dict]):
    def into_response(self, value: dict) -> HttpResponse:
        return JsonResponse(value)


@service(provides=IntoResponse[set])
class SetIntoResponse(IntoResponse[set]):
    def into_response(self, value: set) -> HttpResponse:
        return JsonResponse(value)


@service(provides=IntoResponse[tuple])
class TupleIntoResponse(IntoResponse[tuple]):
    def into_response(self, value: tuple) -> HttpResponse:
        return JsonResponse(value)


@service(provides=IntoResponse[None])
class NoneIntoResponse(IntoResponse[None]):
    def into_response(self, value: None) -> HttpResponse:
        return HttpResponse(status=HTTPStatus.NO_CONTENT)



from collections.abc import Awaitable, Callable
from typing import Optional, Protocol

from asgikit.responses import HttpResponse

from selva.web.request import RequestContext


class Middleware(Protocol):
    async def execute(
        self,
        chain: Callable[[RequestContext], Awaitable[Optional[HttpResponse]]],
        context: RequestContext,
    ) -> Optional[HttpResponse]:
        pass



from collections.abc import Awaitable
from typing import Generic, TypeVar

from asgikit.responses import HttpResponse

__all__ = ("IntoResponse",)

T = TypeVar("T")


class IntoResponse(Generic[T]):
    def into_response(self, value: T) -> HttpResponse | Awaitable[HttpResponse]:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['selva',
 'selva.di',
 'selva.di.service',
 'selva.utils',
 'selva.web',
 'selva.web.middleware',
 'selva.web.request',
 'selva.web.response',
 'selva.web.routing']

package_data = \
{'': ['*']}

install_requires = \
['asgikit>=0.1.2,<0.2.0']

extras_require = \
{':python_version < "3.11"': ['tomli>=2.0.0,<3.0.0']}

setup_kwargs = {
    'name': 'selva',
    'version': '0.1.0',
    'description': 'ASGI Web Framework with Dependency Injection',
    'long_description': '# Project Selva\n\nSelva is a Python ASGI web framework built on top of [asgikit](https://github.com/livioribeiro/asgikit)\nand inspired by Spring Boot, AspNet Core and FastAPI.\n\n## Usage\n\nCreate an application and controller\n\n```python\nfrom selva.web import Application, controller, get\n\n\n@controller\nclass Controller:\n    @get\n    def hello(self):\n        return "Hello, World!"\n\n\napp = Application(Controller)\n```\n\nAdd a service\n\n```python\nfrom selva.di import service\nfrom selva.web import Application, controller, get\n\n\n@service\nclass Greeter:\n    def greet(self, name: str) -> str:\n        return f"Hello, {name}!"\n\n\n@controller\nclass Controller:\n    def __init__(self, greeter: Greeter):\n        self.greeter = greeter\n\n    @get\n    def hello(self):\n        return self.greeter.greet("World")\n\n\napp = Application(Greeter, Controller)\n```\n\nGet parameters from path\n\n```python\nfrom selva.di import service\nfrom selva.web import Application, controller, get\n\n\n@service\nclass Greeter:\n    def greet(self, name: str) -> str:\n        return f"Hello, {name}!"\n\n\n@controller\nclass Controller:\n    def __init__(self, greeter: Greeter):\n        self.greeter = greeter\n\n    @get("hello/{name}")\n    def hello(self, name: str):\n        greeting = self.greeter.greet(name)\n        # A json response will be created from the returned dict\n        return {"greeting": greeting}\n\n\napp = Application(Greeter, Controller)\n```\n\nConfigurations with [Pydantic](https://pydantic-docs.helpmanual.io/usage/settings/)\n\n```python\nfrom selva.di import service\nfrom selva.web import Application, RequestContext, controller, get\nfrom pydantic import BaseSettings\n\n\nclass Settings(BaseSettings):\n    DEFAULT_NAME = "World"\n\n\n@service\ndef settings_factory() -> Settings:\n    return Settings()\n\n\n@service\nclass Greeter:\n    def __init__(self, settings: Settings):\n        self.default_name = settings.DEFAULT_NAME\n\n    def greet(self, name: str | None) -> str:\n        name = name or self.default_name\n        return f"Hello, {name}!"\n\n\n@controller\nclass Controller:\n    def __init__(self, greeter: Greeter):\n        self.greeter = greeter\n\n    @get("hello/{name}")\n    def hello(self, name: str):\n        greeting = self.greeter.greet(name)\n        return {"greeting": greeting}\n\n    @get("hello")\n    def hello_optional(self, context: RequestContext):\n        name = context.query.get("name")\n        greeting = self.greeter.greet(name)\n        return {"greeting": greeting}\n\n\napp = Application(Greeter, Controller)\n```\n\nManage services lifecycle (e.g [Databases](https://www.encode.io/databases/))\n\n```python\nfrom selva.di import service, initializer, finalizer\nfrom selva.web import Application, RequestContext, controller, get\nfrom pydantic import BaseSettings, PostgresDsn\nfrom databases import Database\n\n\nclass Settings(BaseSettings):\n    DEFAULT_NAME = "World"\n    DATABASE_URL: PostgresDsn\n\n\n@service\ndef settings_factory() -> Settings:\n    return Settings()\n\n\n@service\nclass Repository:\n    def __init__(self, settings: Settings):\n        self.database = Database(settings.DATABASE_URL)\n\n    async def get_greeting(self, name: str) -> str:\n        result = await self.database.fetch_one(\n            query="select text from greeting where name = :name",\n            values={"name": name}\n        )\n\n        return result.text\n\n    @initializer\n    async def initialize(self):\n        await self.database.connect()\n        print("Database connection opened")\n\n    @finalizer\n    async def finalize(self):\n        await self.database.disconnect()\n        print("Database connection closed")\n\n\n@service\nclass Greeter:\n    def __init__(self, repository: Repository, settings: Settings):\n        self.repository = repository\n        self.default_name = settings.DEFAULT_NAME\n\n    async def greet(self, name: str | None) -> str:\n        name = name or self.default_name\n        return await self.repository.get_greeting(name)\n\n\n@controller\nclass Controller:\n    def __init__(self, greeter: Greeter):\n        self.greeter = greeter\n\n    @get("hello/{name}")\n    def hello(self, name: str):\n        greeting = self.greeter.greet(name)\n        return {"greeting": greeting}\n\n    @get("hello")\n    def hello_optional(self, context: RequestContext):\n        name = context.query.get("name")\n        greeting = self.greeter.greet(name)\n        return {"greeting": greeting}\n\n\napp = Application(settings_factory, Repository, Greeter, Controller)\n```\n\nDefine controllers and services in a separate module\n\n```\n├───application\n│   ├───controllers.py\n│   ├───repository.py\n│   ├───services.py\n│   └───settings.py\n└───main.py\n```\n\n```python\n### application/settings.py\nfrom selva.di import service\nfrom pydantic import BaseSettings, PostgresDsn\n\n\nclass Settings(BaseSettings):\n    DEFAULT_NAME = "World"\n    DATABASE_URL: PostgresDsn\n\n\n@service\ndef settings_factory() -> Settings:\n    return Settings()\n```\n\n```python\n### application/repository.py\nfrom selva.di import service, initializer, finalizer\nfrom databases import Database\nfrom .settings import Settings\n\n@service\nclass Repository:\n    def __init__(self, settings: Settings):\n        self.database = Database(settings.DATABASE_URL)\n\n    async def get_greeting(self, name: str) -> str:\n        result = await self.database.fetch_one(\n            query="select text from greeting where name = :name",\n            values={"name": name}\n        )\n\n        return result.text\n\n    @initializer\n    async def initialize(self):\n        await self.database.connect()\n        print("Database connection opened")\n\n    @finalizer\n    async def finalize(self):\n        await self.database.disconnect()\n        print("Database connection closed")\n```\n\n```python\n### application/services.py\nfrom selva.di import service\nfrom .settings import Settings\nfrom .repository import Repository\n\n\n@service\nclass Greeter:\n    def __init__(self, repository: Repository, settings: Settings):\n        self.repository = repository\n        self.default_name = settings.DEFAULT_NAME\n\n    async def greet(self, name: str | None) -> str:\n        name = name or self.default_name\n        return await self.repository.get_greeting(name)\n```\n\n```python\n### application/controllers.py\nfrom selva.web import RequestContext, controller, get\nfrom .services import Greeter\n\n@controller\nclass Controller:\n    def __init__(self, greeter: Greeter):\n        self.greeter = greeter\n\n    @get("hello/{name}")\n    def hello(self, name: str):\n        greeting = self.greeter.greet(name)\n        return {"greeting": greeting}\n\n    @get("hello")\n    def hello_optional(self, context: RequestContext):\n        name = context.query.get("name")\n        greeting = self.greeter.greet(name)\n        return {"greeting": greeting}\n```\n\n```python\n### main.py\nfrom selva.web import Application\n\n\n# module named "application" is automatically registered\napp = Application()\n```\n\nWebsockets\n\n```python\nfrom pathlib import Path\nfrom selva.web import Application, FileResponse, RequestContext, controller, get, websocket\nfrom selva.web.errors import WebSocketDisconnectError\n\n@controller\nclass WebSocketController:\n    @get\n    def index(self) -> FileResponse:\n        return FileResponse(Path(__file__).parent / "index.html")\n\n    @websocket("/chat")\n    async def chat(self, context: RequestContext):\n        client = context.websocket\n\n        await client.accept()\n        print(f"[open] Client connected")\n\n        self.handler.clients.append(client)\n\n        while True:\n            try:\n                message = await client.receive()\n                print(f"[message] {message}")\n                await client.send_text(message)\n            except WebSocketDisconnectError:\n                print("[close] Client disconnected")\n                break\n\n\napp = Application(WebSocketController)\n```\n\n```html\n<!DOCTYPE html>\n<html lang="en">\n<head>\n    <meta charset="UTF-8">\n    <title>WebSocket chat</title>\n</head>\n<body>\n<form id="chat-form">\n    <textarea name="message-list" id="message-list" cols="30" rows="10" readonly></textarea>\n    <p>\n        <input type="text" name="message-box" id="message-box" />\n        <button type="submit">Send</button>\n    </p>\n</form>\n\n<script>\n    const messages = [];\n\n    const chat = document.getElementById("chat-form");\n    const textarea = document.getElementById("message-list");\n    textarea.value = "";\n    const messageInput = document.getElementById("message-box");\n\n    const socket = new WebSocket("ws://localhost:8000/chat");\n\n    function addMessage(message) {\n        messages.push(message)\n        textarea.value = `${messages.join("\\n")}`;\n        textarea.scrollTop = textarea.scrollHeight;\n    }\n\n    chat.onsubmit = (event) => {\n        event.preventDefault();\n        const message = messageInput.value;\n        socket.send(message);\n        messageInput.value = "";\n    };\n\n    socket.onopen = (event) => {\n        console.log("[open] Client connected");\n    };\n\n    socket.onmessage = (event) => {\n        const message = event.data;\n        console.log(`[message] "${message}"`)\n        addMessage(message);\n    };\n\n    socket.onclose = (event) => {\n        if (event.wasClean) {\n            console.log(`[close] Connection closed cleanly, code=${event.code} reason=${event.reason}`);\n        } else {\n            console.log(\'[close] Connection died\');\n        }\n    };\n\n    socket.onerror = function(error) {\n        console.log(`[error] ${error.message}`);\n    };\n</script>\n</body>\n</html>\n```\n',
    'author': 'Livio Ribeiro',
    'author_email': 'livioribeiro@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/livioribeiro/selva',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import asyncio
import inspect
from collections.abc import Awaitable, Callable, Coroutine
from typing import Any


async def maybe_async(target: Awaitable | Callable | Coroutine, *args, **kwargs) -> Any:
    if inspect.isawaitable(target):
        return await target

    if inspect.iscoroutinefunction(target):
        return await target(*args, **kwargs)

    return await asyncio.to_thread(target, *args, **kwargs)



from typing import Generic, TypeVar

from .context import RequestContext

__all__ = ("FromRequest",)

T = TypeVar("T")


class FromRequest(Generic[T]):
    async def from_request(self, context: RequestContext) -> T:
        pass



from http import HTTPStatus

from asgikit.errors.websocket import WebSocketDisconnectError, WebSocketStateError


class HttpError(Exception):
    def __init__(self, status: int | HTTPStatus):
        super().__init__()
        if status < 400:
            raise ValueError(
                f"HttpError status should be client or server error, {status} given"
            )
        self.status = status


class ClientError(HttpError):
    pass


class NotFoundError(ClientError):
    def __init__(self):
        super().__init__(HTTPStatus.NOT_FOUND)


class UnauthorizedError(ClientError):
    def __init__(self):
        super().__init__(HTTPStatus.UNAUTHORIZED)


class ForbidenError(ClientError):
    def __init__(self):
        super().__init__(HTTPStatus.FORBIDDEN)


class ServerError(HttpError):
    pass


class InternalServerError(ServerError):
    def __init__(self):
        super().__init__(HTTPStatus.INTERNAL_SERVER_ERROR)



import functools
from collections import defaultdict
from collections.abc import Callable
from types import ModuleType
from typing import Any, Optional, TypeVar, Union

from selva.utils.maybe_async import maybe_async
from selva.utils.package_scan import scan_packages

from .decorators import DI_SERVICE_ATTRIBUTE
from .errors import (
    DependencyLoopError,
    InstanceNotDefinedError,
    InvalidScopeError,
    MissingDependentContextError,
    ServiceNotFoundError,
    ServiceWithoutDecoratorError,
)
from .service.model import InjectableType, Scope, ServiceDefinition, ServiceDependency
from .service.parse import get_dependencies, parse_definition
from .service.registry import ServiceRegistry

TService = TypeVar("TService")


class Container:
    def __init__(self):
        self.registry = ServiceRegistry()
        self.store_singleton: dict[type, Any] = {}
        self.store_dependent: dict[int, dict[type, Any]] = defaultdict(dict)
        self.finalizers: dict[int | None, list[Callable]] = defaultdict(list)

    def register(
        self,
        service: InjectableType,
        scope=Scope.SINGLETON,
        *,
        provides: type = None,
        name: str = None,
    ):
        definition = parse_definition(service, scope, provides)
        provided_service = definition.provides
        self.registry[provided_service, name] = definition

    def service(self, service_type: type):
        service_info = getattr(service_type, DI_SERVICE_ATTRIBUTE, None)
        if not service_info:
            raise ServiceWithoutDecoratorError(service_type)

        scope, provides, name = service_info
        self.register(service_type, scope, provides=provides, name=name)

    def define_singleton(self, service_type: type, instance: Any):
        self.store_singleton[service_type] = instance

    def define_dependent(self, service_type: type, instance: Any, *, context: Any):
        if context is None:
            raise MissingDependentContextError()

        if not self.has(service_type, Scope.DEPENDENT):

            def defined_only_factory() -> service_type:
                raise InstanceNotDefinedError(service_type)

            self.register(defined_only_factory, scope=Scope.DEPENDENT)

        # service = weakref.proxy(instance) if instance is context else instance
        self.store_dependent[id(context)][service_type] = instance

    def scan(self, *packages: Union[str, ModuleType]):
        def predicate(item: Any):
            return hasattr(item, DI_SERVICE_ATTRIBUTE)

        for service in scan_packages(packages, predicate):
            scope, provides, name = getattr(service, DI_SERVICE_ATTRIBUTE)
            self.register(service, scope, provides=provides, name=name)

    def has(self, service: type, scope: Scope = None) -> bool:
        definition = self.registry.get(service)

        if not definition:
            return False

        if scope:
            return definition.scope == scope

        return True

    async def get(
        self,
        service_type: TService,
        *,
        context: Any = None,
        name: str = None,
        optional=False,
    ) -> TService:
        dependency = ServiceDependency(service_type, name=name, optional=optional)
        return await self._get(dependency, context)

    async def create(self, service: type, *, context: Any = None) -> Any:
        dependencies = await self._resolve_dependencies(service, context)
        instance = service(**dependencies)
        return instance

    async def _resolve_dependencies(self, service: InjectableType, context: Any = None):
        return {
            name: await self._get(dep, context)
            for name, dep in get_dependencies(service)
        }

    async def _get(
        self,
        dependency: ServiceDependency,
        context: Any = None,
        valid_scope: Scope = None,
        stack: list = None,
    ) -> Optional[Any]:
        service_type, name = dependency.service, dependency.name

        # try getting from singleton store
        if instance := self.store_singleton.get(service_type):
            return instance

        # try from the dependent store
        if (store := self.store_dependent.get(id(context))) and (
            instance := store.get(service_type)
        ):
            return instance

        try:
            definition = self.registry[service_type, name]
        except ServiceNotFoundError:
            if dependency.optional:
                return None
            raise

        valid_scope = valid_scope or definition.scope

        if not definition.accept_scope(valid_scope):
            requester = stack[0] if stack else None
            raise InvalidScopeError(
                service_type, definition.scope.name, valid_scope.name, requester
            )

        if definition.scope is Scope.DEPENDENT and context is None:
            raise MissingDependentContextError()

        stack = stack or []

        if service_type in stack:
            raise DependencyLoopError(stack + [service_type])

        stack.append(service_type)

        instance = await self._create_service(definition, valid_scope, stack, context)

        return instance

    async def _create_service(
        self,
        definition: ServiceDefinition,
        valid_scope: Scope,
        stack: list[type],
        context: Any = None,
    ) -> Any:
        factory = definition.factory

        dependencies = {
            name: await self._get(dep, context, valid_scope, stack)
            for name, dep in definition.dependencies
        }

        instance = await maybe_async(factory, **dependencies)

        match definition.scope:
            case Scope.SINGLETON:
                self.store_singleton[definition.provides] = instance
            case Scope.DEPENDENT:
                self.store_dependent[id(context)][definition.provides] = instance

        await self._run_initializer(definition, instance, context)
        await self._setup_finalizer(definition, instance, context)

        return instance

    async def _run_initializer(
        self, definition: ServiceDefinition, instance: Any, context: Any | None
    ):
        initializer = definition.initializer

        if initializer is None:
            return

        dependencies = await self._resolve_dependencies(initializer, context)
        await maybe_async(initializer, instance, **dependencies)

    async def _setup_finalizer(
        self, definition: ServiceDefinition, instance: Any, context: Any | None
    ):
        finalizer = definition.finalizer

        if finalizer is None:
            return

        finalizer_key = None if definition.scope is Scope.SINGLETON else id(context)
        self.finalizers[finalizer_key].append(functools.partial(finalizer, instance))

    async def run_finalizers(self, context=None):
        context_id = id(context) if context else None

        for finalizer in self.finalizers[context_id]:
            await maybe_async(finalizer)

        self.finalizers.pop(context_id, None)

        if context_id:
            self.store_dependent.pop(context_id, None)



from selva.di import service

from .converter import PathParamConverter


@service(provides=PathParamConverter[int])
class IntPathParamConverter(PathParamConverter[int]):
    async def from_path(self, value: str) -> int:
        return int(value)

    async def to_path(self, obj: int) -> str:
        return str(obj)


@service(provides=PathParamConverter[float])
class FloatPathParamConverter(PathParamConverter[float]):
    async def from_path(self, value: str) -> float:
        return float(value)

    async def to_path(self, obj: float) -> str:
        return str(obj)


@service(provides=PathParamConverter[str])
class StrPathParamConverter(PathParamConverter[str]):
    async def from_path(self, value: str) -> str:
        return value

    async def to_path(self, obj: str) -> str:
        return obj


@service(provides=PathParamConverter[bool])
class BoolPathParamConverter(PathParamConverter[bool]):
    async def from_path(self, value: str) -> bool:
        return value in ["1", "true", "True"]

    async def to_path(self, obj: bool) -> str:
        return str(obj)



import inspect
from collections.abc import Callable
from typing import Any

from .service.model import InjectableType


def _type_name(service):
    if (
        inspect.isclass(service)
        or inspect.isfunction(service)
        or inspect.ismethod(service)
    ):
        return service.__qualname__

    return str(service)


class DependencyInjectionError(Exception):
    pass


class InvalidScopeError(DependencyInjectionError):
    def __init__(
        self, service: type, scope: str, requested_scope: str, requester: type = None
    ):
        msg = f"service '{_type_name(service)}' with scope '{scope}'"

        if requester:
            msg += f", requested from '{_type_name(service)}'"

        msg += f" cannot be requested from scope '{requested_scope}'."

        super().__init__(msg)


class DependencyLoopError(DependencyInjectionError):
    def __init__(self, dependency_stack: list[type]):
        loop = " -> ".join([dep.__name__ for dep in dependency_stack])
        super().__init__(f"dependency loop detected: {loop}")


class IncompatibleTypesError(DependencyInjectionError):
    def __init__(self, implementation: type, interface: type):
        super().__init__(
            f"service '{_type_name(implementation)}'"
            f" does not derive from '{_type_name(interface)}'"
        )


class ServiceNotFoundError(DependencyInjectionError):
    def __init__(
        self,
        service: type,
        name: str = None,
    ):
        message = f"unable to find service '{_type_name(service)}'"
        if name is not None:
            message += f" with name '{name}'"

        super().__init__(message)


class MissingDependentContextError(DependencyInjectionError):
    def __init__(self):
        super().__init__("context is required for DEPENDENT scope")


class NonInjectableTypeError(DependencyInjectionError):
    def __init__(self, service: Any):
        super().__init__(f"'{service}' is not injectable")


class FactoryMissingReturnTypeError(DependencyInjectionError):
    def __init__(self, factory: Callable):
        super().__init__(f"factory '{_type_name(factory)}' is missing return type")


class ServiceAlreadyRegisteredError(DependencyInjectionError):
    def __init__(self, service: type, name: str = None):
        message = f"service '{_type_name(service)}' is already registered"
        if name:
            message += f" with name '{name}'"

        super().__init__(message)


class CalledNonCallableError(DependencyInjectionError):
    def __init__(self, called: Any):
        super().__init__(f"{_type_name(called)} is not callable")


class TypeVarInGenericServiceError(DependencyInjectionError):
    def __init__(self, provided: type):
        super().__init__(f"{_type_name(provided)} has generic types")


class MultipleNameAnnotationsError(DependencyInjectionError):
    def __init__(self, names: list[str], parameter: str, service: InjectableType):
        super().__init__(
            f"multiple 'Name' annotations"
            f" for parameter '{parameter}"
            f" on service '{_type_name(service)}':"
            f" {', '.join(names)}"
        )


class InvalidServiceTypeError(DependencyInjectionError):
    def __init__(self, service: Any):
        super().__init__(f"object of type {type(service)} is not a valid service")


class ServiceWithoutDecoratorError(DependencyInjectionError):
    def __init__(self, service: InjectableType):
        super().__init__(
            f"service {service.__module__}.{service.__qualname__}"
            " must be decorated with @service"
        )


class InstanceNotDefinedError(DependencyInjectionError):
    def __init__(self, service: type):
        super().__init__(
            f"{service.__module__}.{service.__qualname__} must be defined"
            " for the given context"
        )
